import pandas as pd
import numpy as np
from skimage.morphology import binary_dilation, binary_erosion, dilation, erosion, disk
from skimage.measure import label, regionprops
from yeastvision.models.unet import unet
from yeastvision.track.cell import getBirthFrame, getLifeData
import matplotlib.pyplot as plt 
from yeastvision.track.utils import normalize_dict_by_sum
from tqdm import tqdm
from skimage.morphology import skeletonize

def count_objects(labeledMask):
    return np.count_nonzero(np.unique(labeledMask))

def get_daughters(tree, cell):
    potentialDaughters = tree[cell-1]
    daughters = list(np.where(potentialDaughters)[0]+1)
    return daughters

def has_daughters(tree, cell):
    return np.any(tree[cell-1,:])   

    
def get_generations(tree, initial_generation):
    generations = [[]]
    for val in initial_generation:
        generations[0].append(val)
        get_generations_recur(tree, 1, generations, 0, val)
    print(generations)
    return generations

def get_generations_recur(tree, generation, generations, max_generation, cell):
    if not has_daughters(tree, cell):
        return 
    
    daughters = get_daughters(tree, cell)
    if generation>=len(generations):
        generations.append([])
        max_generation = generation
    for daughter in daughters:
        generations[generation].append(daughter)
        get_generations_recur(tree, generation+1, generations, max_generation, daughter)

def get_gametes(skeleton, cyto, idx):
    found_gametes = False
    overlap = []
    previous_overlaps = []
    while (not found_gametes) and idx>=0:
        mask = cyto[idx]
        overlap = np.unique(mask[skeleton>0])
        overlap = overlap[overlap!=0]

        previous_overlaps.append(overlap)
        found_gametes = len(overlap)==2
        idx -=1
    
    if len(overlap)==0:
        for previous_overlap in previous_overlaps:
            if len(previous_overlap)>0:
                overlap = previous_overlap
                break

    if len(overlap)>2:
        overlap = overlap[0:2]
    else:
        overlap = np.append(overlap, [None, None])
    return found_gametes, overlap

def is_mating_cell(mat, cyto, cell):
    return np.all(mat[cyto==cell]==cell)

def compute_mating_lineage(tracked, cyto):

    cellVals = np.unique(cyto)
    cellVals = cellVals[cellVals!=0]

    matingCells = np.unique(tracked)
    matingCells = matingCells[matingCells!=0]

    numCells = np.count_nonzero(matingCells)
    
    print("computing lineages for", numCells, "cells")
    gamete_dict = {
                  "gamete1":[],
                    "gamete2":[],
                    "isMating":[],
                    "foundGametes":[]}

    for label in tqdm(cellVals):
        isMating = label in matingCells
        if isMating:
            birth = getBirthFrame(tracked, label)
            mat_birth_mask = (tracked[birth] == label).astype(np.uint8)
            found_gametes, gametes = get_gametes(skeletonize(mat_birth_mask),
                                                 cyto, birth-1)
        else:
            found_gametes = False
            gametes = [None, None]
        
        gamete_dict["gamete1"].append(gametes[0])
        gamete_dict["gamete2"].append(gametes[1])
        gamete_dict["isMating"].append(isMating)
        gamete_dict["foundGametes"].append(found_gametes)
    
    return pd.DataFrame.from_dict(gamete_dict)

class LineageConstruction():
    def __init__(self, segMasks, budMasks = None, backskip = 0, forwardskip = 0):
        self.cellMasks = segMasks 
        self.neckMasks = budMasks
        self.backskip = backskip
        self.forskip = forwardskip
        self.maxT = len(self.cellMasks)-1
        self.budNET = None
        
        self.labelMasks()

        # print("--shaving masks--")
        # self.shaveMasks()

        # self.labelMasks()
    

    
    def shaveMasks(self):
        i = 0
        for cellmask, neckmask in zip(self.cellMasks, self.neckMasks):
            print("\tshave mask", i)
            newNeckMask = np.zeros_like(neckmask)
            if np.any(neckmask):
                for neckI in np.unique(neckmask):
                    if neckI>0:
                        print("\t\t", neckI)
                        newNeckMask += self.processNeckMask(cellmask, neckmask==neckI)
            self.neckMasks[i] = newNeckMask      
            i+=1


    def processNeckMask(self, segIm, neckMask):
        processCount = 0
        overlapCount = count_objects(segIm[neckMask])
        r = 1
        newMask = neckMask.copy()
        footprint =  disk(r)
        while overlapCount != 2 and processCount<2:
            if overlapCount>2:
                newMask = binary_erosion(newMask, footprint = footprint)
            elif overlapCount<2:
                newMask = binary_dilation(newMask, footprint = footprint)
            overlapCount = count_objects(segIm[newMask])
            processCount+=1
        
        if overlapCount !=2:
            return neckMask.astype(np.uint8)
        else:
            return newMask
        
            
    def labelMasks(self):
        self.neckMasks = [label((mask>0).astype(np.uint16)) for mask in self.neckMasks]

    def getInitialCells(self):
        if count_objects(self.cellMasks[0])>1:
                self.initialCells = np.unique(self.neckMasks[0])
                self.initialCells = self.initialCells[self.initialCells!=0]

                if count_objects(self.neckMasks[0])>0:
                    pass
        else:
            self.initialCells = None

    def computeLineages(self):
        numCells = np.count_nonzero(np.unique(self.cellMasks))
        cellVals = np.unique(self.cellMasks)
        cellVals = cellVals[cellVals!=0]

        print("computing lineages for", numCells, "cells")
        daughterArray = np.zeros((numCells, numCells), dtype = bool) # rows = cell, column = daughter
        motherDict = getLifeData(self.cellMasks)
        motherDict["cell"] = cellVals
        motherDict["mother"] = []
        motherDict["confidence"] = []

        for cellId in tqdm(cellVals):
            mother, confidence = self.getMother(cellId)
            if mother:
                daughterArray[mother-1, cellId-1] = True
            motherDict["mother"].append(mother)
            motherDict["confidence"].append(confidence)
        
        motherDF = pd.DataFrame.from_dict(motherDict)
        #daughterDF = pd.DataFrame(data = daughterArray, index = cellVals, columns=cellVals)
        #motherDF = motherDF.set_index("cell")
        return daughterArray, motherDF[["birth", "death", "mother", "confidence"]]
    
    def getMother(self, cellNum):
        '''
        1) get bud that overlaps with daughter the most
        2) '''
        birth = getBirthFrame(self.cellMasks, cellNum)
        birthFrame = self.cellMasks[birth]
        eligibleFrames = slice(max(0, birth-self.backskip), min(self.maxT+1, birth+self.forskip))


        overlapcounts = {}
        i = 0
        for cellframe, budframe in zip(self.cellMasks[eligibleFrames], self.neckMasks[eligibleFrames]):
            if i < birth:
                cellframe = self.cellMasks[birth]

            potentialBuds, counts= np.unique(budframe[cellframe==cellNum], return_counts=True)
            if np.any(potentialBuds>0):
                potentialBuds, counts = potentialBuds[potentialBuds!=0], counts[potentialBuds!=0]
                correctBud = potentialBuds[np.argmax(counts)]
                mothers, counts = np.unique(cellframe[budframe==correctBud], return_counts=True)
                for mother, count in zip(mothers, counts):
                    if mother !=cellNum and mother!=0:
                        if mother in overlapcounts:
                            overlapcounts[mother] += count
                        else:
                            overlapcounts[mother] = count  
            i+=1   
        if overlapcounts:
            overlapcounts = normalize_dict_by_sum(overlapcounts)
            mother = max(overlapcounts, key = overlapcounts.get)
            confidence = overlapcounts[mother]

            if birth==0:
                motherArea = np.sum(birthFrame == mother)
                daughterArea = np.sum(birthFrame==cellNum)
                if daughterArea > motherArea:
                    return None, 0
            return mother, round(confidence,2)
        else:
            return None, 0














    



    
    









import os 
from skimage.io import imread
import glob
from yeastvision.utils import get_mask_contour
import numpy as np
from .utils import *
import copy
from tqdm import tqdm
from yeastvision.track.data import LineageData, TimeSeriesData



class Experiment():

    def __init__(self, dir, num_channels = 1):
        self.dir = dir
        print(self.dir)
        print(os.path.split(dir))
        _, self.name = os.path.split(self.dir)
        self.num_channels = num_channels
        self.num_labels = 0

        self.channels = []
        self.labels = []

        im_npzs, mask_npzs = get_im_mask_npzs(self.dir)
        if mask_npzs:
            self.load_with_mask_npzs(mask_npzs, num_channels)
        else:
            self.load_from_dir(num_channels)
        
        if im_npzs:
            self.load_channels_from_npz_files(im_npzs)
        
        self.set_mask_type("labels")
    
    def new_channel_name(self, index, new_name):
        if new_name not in self.get_channel_names():
            self.channels[index].set_name(new_name)
            return True
        else:
            return False

    def new_label_name(self, index, new_name):
        if new_name not in self.get_channel_names():
            self.labels[index].set_name(new_name)
            return True
        else:
            return False
    
    def load_channels_from_npz_files(self, npz_channel_files):
        self.npz_channel_files = npz_channel_files
        for path in self.npz_channel_files:
            _, name = os.path.split(path)
            if InterpolatedChannel.text_id in name:
                self.add_channel_object(InterpolatedChannel(npz_path=path))
            else:
                self.add_channel_object(ChannelNoDirectory(npz_path=path))
    
    def get_channel_names(self):
        return [channel.name for channel in self.channels]
    
    def get_label_names(self):
        return [label.name for label in self.labels]
        
    def get_dummy_labels(self, full = True):
        shape = self.shape()
        if full:
            h,w = shape
            t = self.get_t()
            return np.zeros((t,h,w), dtype = np.uint8)
        else:
            return np.zeros(shape, dtype = np.uint8)
    
    def get_label(self, mask_type, name = None, idx = None, t = None):
        if not self.has_labels():
            return self.get_dummy_labels(full = t == None)
        else:
            if mask_type!= self.mask_type:
                self.set_mask_type(mask_type)
            if name:
                idx = self.get_mask_index_from_name(name)
            if t is not None:
                return self.labels[idx].data[t,:,:]
            else:
                return self.labels[idx].data[:,:,:]
    
    def get_channel_by_name(self, name):
        for channel in self.channels:
            if channel.name == name:
                return channel
        raise IndexError

    def get_label_by_name(self,name):
        for label in self.labels:
            if label.name == name:
                return label
        raise IndexError
    
    def get_channel(self, name = None, idx = None, t = None):
        if name:
            channel = self.get_channel_by_name(name)
        else:
            channel = self.channels[idx]
        if t is not None:
            return channel.ims[t]
        else:
            return channel.ims
        
    def get_mask_index_from_name(self, name):
        for i,label in enumerate(self.labels):
            if label.name == name:
                return i
        raise IndexError

    def set_mask_type(self, mask_type):
        self.mask_type = mask_type
        for label in self.labels:
            label.set_mask_type(mask_type)
    
    def has_labels(self):
        return self.num_labels>0

    def load_from_dir(self, num_channels):
        channels = [[] for i in range(num_channels)]
        
        files_image_only = sorted(get_image_files(self.dir, remove_masks=True))
        files = sorted(get_image_files(self.dir))

        num_images = len(files_image_only)/num_channels
        num_labels = int(len(files)/num_images - num_channels)

        labels = [[] for i  in range(int(num_labels))]

        groupsize = num_channels+num_labels
        
        for i in range(0, len(files), groupsize):
            group = files[i:i+groupsize]
            channels_in_group = group[0:num_channels]
            labels_group = group[num_channels:groupsize]
            for channel, channel_in_group in zip(channels, channels_in_group):
                channel.append(channel_in_group)
            for label, label_in_group in zip(labels, labels_group):
                label.append(label_in_group)
        
        for channel in channels:
            self.add_channel(channel, increment_count=False)
        
        if num_labels>0:
            for label in labels:
                self.add_label(files = label, increment_count=True)

    def load_with_mask_npzs(self, mask_npzs, num_channels):
        channels = [[]*num_channels]
        
        files = sorted(get_image_files(self.dir, remove_masks=True))
        groupsize = num_channels
        
        for i in range(0, len(files), groupsize):
            group = files[i:i+groupsize]
            channels_in_group = group[0:num_channels]
            labels_group = group[num_channels:groupsize]
            for channel, channel_in_group in zip(channels, channels_in_group):
                channel.append(channel_in_group)
        
        for channel in channels:
            self.add_channel(channel, increment_count=False)
        for npz_path in tqdm(mask_npzs):
            print(npz_path)
            self.labels.append(Label(npz_path=npz_path))
            self.num_labels+=1


    def create_label_objects_from_npz(self):
        self.npzdata = np.load(self.npz_path)
        loaded_npz = self.npzdata
        num_masks = loaded_npz["labels"].shape[0]
        for i in range(num_masks):
            self.labels.append(Label(dir = self.dir, name = self.get_label_name(name = None), npz_name=self.npz_filename, loaded_npz=loaded_npz, npz_index=i))
            self.num_labels+=1
    
    def add_channel_object(self, channel_object):
        self.channels.append(channel_object)
        self.num_channels+=1
    
    def add_label_object(self, label_object):
        self.labels.append(label_object)
        self.num_labels+=1

    def add_channel(self, files, increment_count = True):
        self.channels.append(Channel(files, f"channel_{len(self.channels)}"))

        if increment_count:
            self.num_channels+=1

    def add_label(self, files = None, arrays = None, increment_count = True, name = None, update_data = False):
        new_label = Label(fnames = files, mask_arrays=arrays, dir = self.dir, name = self.get_label_name(name = name))
        self.labels.append(new_label)
        if increment_count:
            self.num_labels +=1
        if update_data:
            self.load_masks()
        
    def get_num_labels(self):
        return len(self.labels)

    def get_num_channels(self):
        return len(self.channels)

    def get_t(self):
        return len(self.channels[0].ims)
    
    def get_label_name(self, name = None):
        if name is None:
            name =  f"label_{len(self.labels)}"
        n = (np.char.find(np.array(self.get_channel_names()), name)+1).sum()
        if n>0:
            return f"{name}-n"
        else:
            return name
    
    def get_channel_name(self, name = None):
        if name is None:
            return f"channel_{len(self.labels)}"
        n = (np.char.find(np.array(self.labels), name)+1).sum()
        if n>0:
            return f"{name}-n"
        else:
            return name
    
    def get_channel_string(self):
        channel_string = ""
        for channel in self.channels:
            channel_string+=f"\t{channel.name}: {str(channel)}\n"
        return channel_string
    
    def get_label_string(self):
        label_string = ""
        for label in self.labels:
            label_string+=f"\t{label.name}: {str(label)}\n"
        return label_string

    def get_viable_im_objs(self, mask_obj):
        viable_ims = []
        for channel in self.channels:
            if channel.t == mask_obj.t and channel.shape == mask_obj.shape:
                viable_ims.append(channel)
        return viable_ims

    def shape(self):
        return self.channels[0].shape
    
    def __str__(self):
        return f"dir: {self.dir}\nt:{self.get_t()}\nnum_channels: {len(self.channels)}\n{self.get_channel_string()}\nnum_labels:{len(self.labels)}\n{self.get_label_string()}"
        


class Files():
    image_extensions = ['.tif','.tiff',
                            '.jpg','.jpeg','.png','.bmp',
                            '.pbm','.pgm','.ppm','.pxm','.pnm','.jp2',
                            '.TIF','.TIFF',
                            '.JPG','.JPEG','.PNG','.BMP',
                            '.PBM','.PGM','.PPM','.PXM','.PNM','.JP2']
    '''
    Lowest level class for storing information about a related set of filenames which should exist in the same directory and should be identifiable by a certain keyword'''
    def __init__(self, fnames, load_on_init = True):
        self.files = fnames
        self.dir = os.path.dirname(self.files[0])
        self.ims = None
        self.loaded = False
        self.load()
        self.get_properties()
        self.annotations = None
    
    def load(self):
        self.ims = [imread(file) for file in self.files]
        self.loaded = True
    
    def deload(self):
        del self.ims
        self.ims = None
        self.loaded = False
    
    def get_properties(self):
        self.t = len(self.ims)
        self.shape = self.ims[0].shape
        self.dtype = str(self.ims[0].dtype)
    
    
    def get_string(self, t):
        return f"{self.shape} {self.dtype}"
    
    def get_files(self, t):
        if t > self.max_t():
            return ""
        return self.files[t]

    def max_t(self):
        return self.t-1

    def set_name(self, new_name):
        self.name = new_name
    
    def x(self):
        return self.shape[-1]

    def y(self):
        return self.shape[0]

    def get_file_names(self):
        return [get_file_name(file) for file in self.files]


class Channel(Files):
    def __init__(self, fnames, name, labels = None, load_on_init = True):
        super().__init__(fnames, load_on_init=load_on_init)
        self.labels = labels
        self.name = name
        self.compute_saturation()
    def __str__(self):
        return f"{self.shape} | {self.dtype}"
    
    def compute_saturation(self):
        self.saturation = [[im.min(), im.max()] for im in self.ims]
    
    def get_saturation(self, t):
        return self.saturation[t]
    

class ChannelNoDirectory(Channel):
    def __init__(self, npz_path = None, ims = None, dir = None, name = None, annotations = None):
        if npz_path is not None:
            self.dir, fname = os.path.split(npz_path)
            self.name = os.path.splitext(fname)[0]
            data = np.load(npz_path, allow_pickle=True)
            self.annotations = data["annotations"]
            self.ims = data["ims"]
            self.path = npz_path
        else:
            self.dir = dir
            self.name = name 
            self.ims = ims
            self.path = self.make_path()
            if annotations is None:
                annotations = [[] for _ in range(len(ims))]
            self.annotations = np.array(annotations, dtype = object)
            self.save_to_npz()
        
        self.get_properties()
        self.compute_saturation()
        self.files = [self.path]

    def make_path(self):
        return os.path.join(self.dir, self.name+".npz")

    def set_name(self, new_name):
        self.name = new_name
        new_npz_name = self.make_path()
        os.rename(self.path, new_npz_name)
        self.path = new_npz_name

    def save_to_npz(self):
        np.savez(self.path, annotations = self.annotations, ims = self.ims)
    
    def get_string(self,t):
        if self.annotations is not None:
            return f"{super().get_string(t)} ,  {(','.join(self.annotations[t])).upper()}"
        else:
            return super().get_string(t)

    def get_files(self, t):
        return self.path
    
    def get_file_names(self):
        return [get_file_name(self.path)]

class InterpolatedChannel(ChannelNoDirectory):
    text_id = "_interp"
    def __init__(self, interpolation = None, npz_path = None, ims = None, dir = None, name = None, annotations = None):
        if npz_path is not None:
            self.dir, fname = os.path.split(npz_path)
            self.name = os.path.splitext(fname)[0]
            data = np.load(npz_path, allow_pickle=True)
            self.annotations = data["annotations"]
            self.ims = data["ims"]
            self.interp_annotations = data["interpolation"]
            self.infer_interpolation()
            self.path = npz_path
        else:
            self.interpolation = interpolation
            self.dir = dir
            self.name = name 
            self.ims = ims
            self.path = os.path.join(self.dir, self.name+".npz")
            print("annotations", annotations)
            if annotations is not None:
                self.annotations = self.extend_annotations(annotations)
            else:
                self.annotations = [[] for _ in range(len(self.ims))]
            self.add_interpolation_to_annotations()
            self.make_interpolation_annotation()
            self.annotations = np.array(self.annotations)
            self.save_to_npz()

        
        self.get_properties()
        self.compute_saturation()
    
    def extend_annotations(self, prev_annotations):
        if type(prev_annotations)  is not list:
            prev_annotations = prev_annotations.tolist()
        new_annotations_blank = [[] for i in range(len(self.ims))]
        for i,prev_ann in enumerate(prev_annotations):
            new_annotations_blank[i] = prev_annotations[i]
        return new_annotations_blank

    def save_to_npz(self):
        np.savez(self.path, annotations = self.annotations, ims = self.ims, interpolation = self.interp_annotations)
    
    def make_interpolation_annotation(self):
        self.interp_annotations = [False for i in range(len(self.ims))]
        for i in range(0, len(self.ims), self.interpolation):
            self.interp_annotations[i] = True

    def add_interpolation_to_annotations(self):
        for i, ann in enumerate(self.annotations):
            if i % self.interpolation == 0:
                self.annotations[i].append("REAL-IMAGE")
            else:
                self.annotations[i].append("INTERPOLATED")

    def infer_interpolation(self):
        num_real = np.count_nonzero(self.interp_annotations)
        total = len(self.interp_annotations)
        self.interpolation = total/num_real
    


class Label(Files):
        mask_types = ["labels", "probability", "contours"]
        def __init__(self, mask_arrays = None, fnames = None, dir = None, npz_path = None, name = None):
            
            self.celldata = None
            self.lineagedata = None
            self.labels, self.contours, self.probability = None, None, None
            
            if npz_path and os.path.exists(npz_path):
                print("loading from npz")
                self.npz_path = npz_path
                self.get_dir_and_name_from_npz_path()
                self.init_from_npz()

            else:   
                self.name = name           
                if fnames is not None:
                    self.files = fnames
                    self.dir = os.path.dirname(self.files[0])
                    raw_arrays = np.array([imread(file) for file in self.files])
                elif mask_arrays is not None:
                    assert(mask_arrays is not None and dir is not None)
                    self.dir = dir
                    self.files = []
                    raw_arrays = np.array(mask_arrays)            
                self.npz_path = os.path.join(self.dir, name+".npz")
                self.extract_probability(raw_arrays)
                self.extract_contours()
                self.get_properties()

                self.write_to_npz()
                self.unload_data_attrs()
            
            self.load()

        def has_cell_data(self):
            return isinstance(self.celldata, TimeSeriesData)
        
        def has_lineage_data(self):
            return isinstance(self.celldata, LineageData)
        
        def get_dir_and_name_from_npz_path(self):
            self.dir, head = os.path.split(self.npz_path)
            self.name,_ = os.path.splitext(head)

        def load(self, mask_type = "labels"):
            self.npzdata = dict(np.load(self.npz_path, allow_pickle=True))
            self.data = self.npzdata[mask_type]
        
        def set_mask_type(self, mask_type):
            assert mask_type in self.mask_types
            self.mask_type = mask_type
            self.data = self.npzdata[self.mask_type]
        
        def save_current_data(self):
            if self.mask_type == "contours":
                self.contours = self.data
            elif self.mask_type == "labels":
                self.labels = self.data
            self.save()


        def init_from_npz(self):
            data = np.load(self.npz_path, allow_pickle=True)
            self.celldata = data["celldata"][0]
            self.lineagedata = data["lineagedata"][0]
            labels = data["labels"][:,:,:]
            self.shape = labels[0].shape
            self.dtype = str(labels[0].dtype)
            self.t = len(labels)
            self.has_probability = np.any(data["probability"][:,:,:]!=0)
            print("has probability:", self.has_probability)
            del labels
            del data

        def get_properties(self):
            self.t = len(self.labels)
            self.shape = self.labels[0].shape
            self.dtype = str(self.labels[0].dtype)
        
        def __str__(self):
            return f"{self.shape} | {self.dtype} | has_probability: {self.has_probability}"

        def write_to_npz(self):
            np.savez(self.npz_path, **self.get_data_dict())

        def unload_data_attrs(self):
            self.labels = None
            self.contours = None
            self.probability = None

        def get_data_dict(self):
            data = {}
            for property_name in ["labels", "contours", "probability"]:
                data[property_name] = getattr(self,f"get_{property_name}")()
            data["celldata"] = np.array([self.celldata])
            data["lineagedata"] = np.array([self.lineagedata])
            return data
        
        def insert(self, new_mask, t, new_prob_im = None):
            self.npzdata["labels"][t] = new_mask
            self.npzdata["contours"][t]  = get_mask_contour(new_mask)
            if self.has_probability:
                self.npzdata["probability"][t] = new_prob_im
            self.save()

        def extract_probability(self, raw_arrays):
            if len(raw_arrays.shape) == 4:
                self.has_probability = True
                self.labels = raw_arrays[0,:,:,:]
                self.probability = raw_arrays[1,:,:,:]
            else:
                self.has_probability = False
                self.labels = raw_arrays
            
        def extract_contours(self):
            self.contours = [get_mask_contour(mask) for mask in tqdm(self.labels)]
        
        def get_labels(self):
            if self.labels is not None:
                return self.labels
            return self.npzdata["labels"]
        
        def get_contours(self):
            if self.contours is not None:
                return self.contours
            return self.npzdata["contours"]

        def get_probability(self):
            if self.has_probability:
                    if self.probability is not None:
                        return self.probability
                    return self.npzdata["probability"]
            else:
                return np.zeros_like(self.get_labels(), dtype = np.uint8)
        
        def get_files(self,t):
            return self.npz_path
        
        def set_name(self, new_name):
            self.name = new_name
            new_npz_name = self.make_path()
            os.rename(self.npz_path, new_npz_name)
            self.npz_path = new_npz_name
        
        def make_path(self):
            return os.path.join(self.dir, self.name+".npz")
        
        def save(self):
            print("saving")
            self.write_to_npz()
            #self.load()
            #self.unload_data_attrs()
        
        def set_data(self, labels, contours):
            self.labels = labels
            self.contours = contours
            self.save()
            self.unload_data_attrs()
        
        def set_contours(self, contours):
            self.contours = contours
            self.save()
            self.unload_data_attrs()



import torch
import torch.nn as nn
import numpy as np
import torch.optim as optim
import itertools
from ..rife_model.warplayer import warp
import torch.nn.functional as F

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

def conv(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
        nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                  padding=padding, dilation=dilation, bias=True),
        nn.PReLU(out_planes)
        )

def deconv(in_planes, out_planes, kernel_size=4, stride=2, padding=1):
    return nn.Sequential(
        torch.nn.ConvTranspose2d(in_channels=in_planes, out_channels=out_planes, kernel_size=4, stride=2, padding=1, bias=True),
        nn.PReLU(out_planes)
        )
            
class Conv2(nn.Module):
    def __init__(self, in_planes, out_planes, stride=2):
        super(Conv2, self).__init__()
        self.conv1 = conv(in_planes, out_planes, 3, stride, 1)
        self.conv2 = conv(out_planes, out_planes, 3, 1, 1)

    def forward(self, x):
        x = self.conv1(x)
        x = self.conv2(x)
        return x
    
c = 16
class Contextnet(nn.Module):
    def __init__(self):
        super(Contextnet, self).__init__()
        self.conv1 = Conv2(3, c)
        self.conv2 = Conv2(c, 2*c)
        self.conv3 = Conv2(2*c, 4*c)
        self.conv4 = Conv2(4*c, 8*c)
    
    def forward(self, x, flow):
        x = self.conv1(x)
        flow = F.interpolate(flow, scale_factor=0.5, mode="bilinear", align_corners=False, recompute_scale_factor=False) * 0.5
        f1 = warp(x, flow)        
        x = self.conv2(x)
        flow = F.interpolate(flow, scale_factor=0.5, mode="bilinear", align_corners=False, recompute_scale_factor=False) * 0.5
        f2 = warp(x, flow)
        x = self.conv3(x)
        flow = F.interpolate(flow, scale_factor=0.5, mode="bilinear", align_corners=False, recompute_scale_factor=False) * 0.5
        f3 = warp(x, flow)
        x = self.conv4(x)
        flow = F.interpolate(flow, scale_factor=0.5, mode="bilinear", align_corners=False, recompute_scale_factor=False) * 0.5
        f4 = warp(x, flow)
        return [f1, f2, f3, f4]
    
class Unet(nn.Module):
    def __init__(self):
        super(Unet, self).__init__()
        self.down0 = Conv2(17, 2*c)
        self.down1 = Conv2(4*c, 4*c)
        self.down2 = Conv2(8*c, 8*c)
        self.down3 = Conv2(16*c, 16*c)
        self.up0 = deconv(32*c, 8*c)
        self.up1 = deconv(16*c, 4*c)
        self.up2 = deconv(8*c, 2*c)
        self.up3 = deconv(4*c, c)
        self.conv = nn.Conv2d(c, 3, 3, 1, 1)

    def forward(self, img0, img1, warped_img0, warped_img1, mask, flow, c0, c1):
        s0 = self.down0(torch.cat((img0, img1, warped_img0, warped_img1, mask, flow), 1))
        s1 = self.down1(torch.cat((s0, c0[0], c1[0]), 1))
        s2 = self.down2(torch.cat((s1, c0[1], c1[1]), 1))
        s3 = self.down3(torch.cat((s2, c0[2], c1[2]), 1))
        x = self.up0(torch.cat((s3, c0[3], c1[3]), 1))
        x = self.up1(torch.cat((x, s2), 1)) 
        x = self.up2(torch.cat((x, s1), 1)) 
        x = self.up3(torch.cat((x, s0), 1)) 
        x = self.conv(x)
        return torch.sigmoid(x)



from skimage.exposure import equalize_adapthist
from skimage.filters import median, gaussian
from skimage.exposure import equalize_adapthist
from skimage import img_as_ubyte, img_as_uint, img_as_float
import numpy as np
import warnings

def do_adapt_hist(ims):
    return [equalize_adapthist(im) for im in ims]

def do_median(ims, kernel_dim):
    kernel = np.ones((kernel_dim, kernel_dim))
    return [median(im,kernel) for im in ims]

def do_gaussian(ims, sigma):
    return [gaussian(im, sigma = sigma) for im in ims]


def rolling_mean(data, window):
    ''' compute the rolling mean of the data over the given window '''

    result = np.full_like(data, np.nan)

    conv = np.convolve(data, np.ones(window)/window, mode='valid')
    result[(len(data) - len(conv))//2: (len(conv) - len(data))//2] = conv

    return result

def calc_brightness(images, sigma=2.5):
    brightness = []
    for image in images:
        mask = np.ones_like(image, dtype=bool)
        if sigma is not None:
            mean = np.mean(image)
            std = np.std(image)
            dist = np.abs(image - mean) / std
            mask[dist > sigma] = False
        brightness.append(np.mean(image[mask]))
    return np.array(brightness)

def scale_image_brightness(image, scale):
    ''' scale image brightness by a factor '''
    adjusted_image = scale * img_as_float(image)
    # handle overflow:
    adjusted_image[adjusted_image >= 1.0] = 1.0

    # catch warning for loosing some accuracy by converting back to int types
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=UserWarning)
        if image.dtype == np.dtype('uint8'):
            adjusted_image = img_as_ubyte(adjusted_image)
        elif image.dtype == np.dtype('uint16'):
            adjusted_image = img_as_uint(adjusted_image)

    return adjusted_image


def deflicker(images, window = 10, sigma=2.5):
    ''' Deflicker images
    Image brightness is scaled to match a rolling mean to avoid flickering
    Parameters
    ----------
    images: list of strings
        Filenames of the images that should be processed
    window: int
        The width of the window for the rolling mean
    outdir: string
        The directory where the adjusted images are saved
    fmt: string
        Output format. One of png, tiff, jpg
    '''

    brightness = calc_brightness(images, sigma=sigma)
    target_brightness = rolling_mean(brightness, window)
    adjusted_ims = []
    for image, b, tb in zip(images, brightness, target_brightness):
        if np.isnan(tb):
            adjusted_image = image.copy()
        else:
            adjusted_image = scale_image_brightness(image, tb / b)
        adjusted_ims.append(adjusted_image)
    return adjusted_ims







import torch
import numpy as np
import torch.nn as nn
import torch.nn.functional as F

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

import torch

def gauss_kernel(size=5, channels=3):
    kernel = torch.tensor([[1., 4., 6., 4., 1],
                           [4., 16., 24., 16., 4.],
                           [6., 24., 36., 24., 6.],
                           [4., 16., 24., 16., 4.],
                           [1., 4., 6., 4., 1.]])
    kernel /= 256.
    kernel = kernel.repeat(channels, 1, 1, 1)
    kernel = kernel.to(device)
    return kernel

def downsample(x):
    return x[:, :, ::2, ::2]

def upsample(x):
    cc = torch.cat([x, torch.zeros(x.shape[0], x.shape[1], x.shape[2], x.shape[3]).to(device)], dim=3)
    cc = cc.view(x.shape[0], x.shape[1], x.shape[2]*2, x.shape[3])
    cc = cc.permute(0,1,3,2)
    cc = torch.cat([cc, torch.zeros(x.shape[0], x.shape[1], x.shape[3], x.shape[2]*2).to(device)], dim=3)
    cc = cc.view(x.shape[0], x.shape[1], x.shape[3]*2, x.shape[2]*2)
    x_up = cc.permute(0,1,3,2)
    return conv_gauss(x_up, 4*gauss_kernel(channels=x.shape[1]))

def conv_gauss(img, kernel):
    img = torch.nn.functional.pad(img, (2, 2, 2, 2), mode='reflect')
    out = torch.nn.functional.conv2d(img, kernel, groups=img.shape[1])
    return out

def laplacian_pyramid(img, kernel, max_levels=3):
    current = img
    pyr = []
    for level in range(max_levels):
        filtered = conv_gauss(current, kernel)
        down = downsample(filtered)
        up = upsample(down)
        diff = current-up
        pyr.append(diff)
        current = down
    return pyr

class LapLoss(torch.nn.Module):
    def __init__(self, max_levels=5, channels=3):
        super(LapLoss, self).__init__()
        self.max_levels = max_levels
        self.gauss_kernel = gauss_kernel(channels=channels)
        
    def forward(self, input, target):
        pyr_input  = laplacian_pyramid(img=input, kernel=self.gauss_kernel, max_levels=self.max_levels)
        pyr_target = laplacian_pyramid(img=target, kernel=self.gauss_kernel, max_levels=self.max_levels)
        return sum(torch.nn.functional.l1_loss(a, b) for a, b in zip(pyr_input, pyr_target))



from PyQt5 import QtWidgets, QtGui, QtCore
from PyQt5.QtWidgets import (QApplication, QRadioButton, QButtonGroup, QSlider, QStyle, QStyleOptionSlider, QMainWindow, QPushButton, QDialog,
                            QDialogButtonBox, QLineEdit, QFormLayout, QMessageBox,
                             QFileDialog, QVBoxLayout, QCheckBox, QFrame, QSpinBox, QLabel, QWidget, QComboBox, QTableWidget, QSizePolicy, QGridLayout, QHBoxLayout)
import pyqtgraph as pg
from pyqtgraph.dockarea.Dock import Dock
from pyqtgraph.dockarea.DockArea import DockArea
from yeastvision.plot.cell_table import TableModel
from yeastvision.plot.types import SingleCellUpdatePlot, HeatMap, PopulationPlot
from yeastvision.track.data import PopulationReplicate, Population
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar


class PlotProperty():
    def __init__(self, parent, plotType, setName, property, populationName = None):
        self.parent = parent
        self.plotType = plotType
        self.labelIdx = self.getLabelIdx(setName)
        self.prop = property
        self.population = populationName
    
    def getPlotName(self):
        labelName = self.getLabelName()
        if self.populationName:
            return "f{labelName}-{self.population}-{self.prop}"
        else:
            return "f{labelName}-{self.prop}"
    
    def getLabelIdx(self, setName):
        return self.parent.labelSelect.items().index(setName)
    
    def getLabelName(self):
        return self.labelSelect.items()[self.labelIdx]
    
    def get_timeseries_obj(self):
        return self.parent.cellData[self.labelIdx]
    
    @staticmethod
    def cluster_property(plot_property_list):
        by_property = {}
        for plot_property_obj in plot_property_list:
            prop = plot_property_obj.prop
            if prop not in by_property:
                by_property[prop] = [plot_property_obj]
            else:
                by_property[prop].append(plot_property_obj)
        return by_property

    @staticmethod
    def cluster_population(plot_property_list):
        by_population = {}
        for plot_property_obj in plot_property_list:
            pop = plot_property_obj.population
            if pop not in by_population:
                by_population[pop] = [plot_property_obj]
            else:
                by_population[pop].append(plot_property_obj)
        return by_population

    

class EvalWindow(QWidget):
    def __init__(self, parent, data_dict, IOUs):
        '''
        params
        data_dict - key: name of the plot (precision, recall, F1, accuracy), values - list of tuple for each set of labels being 
        evaluated where tup1 is the name and tup2 is data vs IOU score
        '''
        super(EvalWindow, self).__init__()
        self.setWindowTitle("Evaluation Window")

        self.figure = plt.figure()
        self.canvas = FigureCanvas(self.figure)
        self.data = data_dict
        print(self.data)

        self.numPlots = len(data_dict)
        self.evalNames = [n for n,_ in list(self.data.values())[0]]
        self.numToEval = len(self.evalNames)
        self.xVals = IOUs
        print(IOUs)

        # Create a table widget to display data
        self.tableLayout = QVBoxLayout()
        self.table_widgets = []

        self.toolbar = NavigationToolbar(self.canvas, self)


        layout = QVBoxLayout()
        layout.addWidget(self.toolbar)
        layout.addWidget(self.canvas)
        #layout.addWidget(self.tableLayout)

        self.setLayout(layout)

        self.setData()
        self.setGeometry(400,400,500,500)
        self.canvas.draw()
        self.show()

    def setData(self):
        i = 0
        for plotType, data in self.data.items():
            self.producePlot(data, plotType, i)
            i+=1

    
    # def produceTable(self, title, data):
    #     table_widget = QTableWidget()
    #     table_widget.setColumnCount(self.numToEval)
    #     table_widget.setRowCount(len(data[0]))

    #     for col, line_data in enumerate(data):
    #         for row, value in enumerate(line_data):
    #             item = QTableWidgetItem(str(value))
    #             table_widget.setItem(row, col, item)

    #     return table_widgets

    
    def producePlot(self, datas, title, col):
        ax = self.figure.add_subplot(2, 2, col + 1)
        for name, data in datas:
            ax.plot(self.xVals, data, 'o-', label = name)
            ax.set_aspect('auto')
            ax.set_xbound(lower = 0.45,upper = 1.05)
            ax.set_xticks([0.5,0.6,0.7,0.8,0.9,1])

            if col == 0:
                ax.legend()
                ax.set_xlabel("IOU Matching Threshold")
        ax.set_title(title)
        plt.tight_layout()




class PlotWindow(QWidget):
    def __init__(self, parent, propDict):
        super().__init__()
        self.parent = parent
        pg.setConfigOption('imageAxisOrder', 'row-major')
        self.hLayout = QHBoxLayout()
        self.setLayout(self.hLayout)
        self.createTable()
        pg.setConfigOption('imageAxisOrder', 'row-major')
        self.cmap = pg.colormap.get("jet", source = 'matplotlib')
        self.setData()
        
        self.propDict = propDict


        if self.hasPlots():
            self.area = DockArea()
            self.hLayout.addWidget(self.area, 7)
            self.docks = {} # key = name of dock, value = (dock object, widget contained in dock)
            if self.hasSingleCellPlots():
                self.addSingleCellPlots()
            if self.hasHeatmaps():
                self.addHeatmaps()
            if self.hasPopulationPlots():
                self.addPopulationsPlots()
            self.setGeometry(500, 100, 1500, 1500)
    
    def hasPlots(self):
        return self.hasSingleCellPlots() or self.hasHeatmaps() or self.hasPopulationPlots()

    def hasSingleCellPlots(self):
        return "single" in self.propDict and self.propDict["single"]
    
    def hasMplPlots(self):
        return self.hasPopulationPlots()
    
    def hasHeatmaps(self):
        return "heatmap" in self.propDict and self.propDict["heatmap"]
    
    def hasPopulationPlots(self):
        return "population" in self.propDict and self.propDict["population"]

    def updateAll(self):
        if self.hasSingleCellPlots():
            self.updateSingleCellPlots()
        if self.hasHeatmaps():
            self.updateHeatMaps()
        if self.hasPopulationPlots():
            self.updatePopulationPlots

    
    def addSingleCellPlots(self):
        self.singleCellPlots = []
        single_cell_prop_obj = self.propDict["single"]
        clustered_by_property = PlotProperty.cluster_property(single_cell_prop_obj)
        for property in clustered_by_property:
            timeseries = [prop.get_timeseries_obj() for prop in clustered_by_property[property]]
            self.singleCellPlots.append(SingleCellUpdatePlot(self.parent, property, timeseries, self.parent.selectedCells))
        
 
        dockName = f"single plots"
        newDock = Dock(dockName, closable = True)
        dockLayout = DockArea()
        newDock.addWidget(dockLayout)
        self.area.addDock(newDock, "bottom")
        

        for i, plot in enumerate(self.singleCellPlots):
            if i == 0:
                position = "bottom"
                relativeTo = None
            else:
                position = "above"
                plotRelativeTo = self.singleCellPlots[i-1]
                relativeTo = self.docks[plotRelativeTo]

            name = f"{plot.property}"
            plotDock = Dock(name, closable = True)
            plotDock.addWidget(plot.getPlotWidget())
            self.docks[plot] = (plotDock)
  
            dockLayout.addDock(plotDock, position, relativeTo)
    
    def addHeatmaps(self):
        self.heatMaps = []
        heatmap_prop_obj = self.propDict["heatmap"]
        clustered_by_property = PlotProperty.cluster_property(heatmap_prop_obj)
        for property, properties in clustered_by_property.items():
            timeseries = [prop.get_timeseries_obj() for prop in properties]
            populations = [prop.population for prop in properties]
            for ts, population in zip(timeseries, populations):
                self.heatMaps.append(HeatMap(self, property, population, ts))
    
        dockName = f"heatmaps"
        newDock = Dock(dockName, closable = True)
        dockLayout = DockArea()
        newDock.addWidget(dockLayout)
        self.area.addDock(newDock, "bottom")

        for i, plot in enumerate(self.heatMaps):
            if i == 0:
                position = "bottom"
                relativeTo = None
            else:
                position = "above"
                plotRelativeTo = self.heatMaps[i-1]
                relativeTo = self.docks[plotRelativeTo]

            name = f"{plot.property}"
            plotDock = Dock(name, closable = True)
            plotDock.addWidget(plot.getPlotWidget())
            self.docks[plot] = (plotDock)

            dockLayout.addDock(plotDock, position, relativeTo)
    
    def addPopulationsPlots(self):
        self.populationPlots = []
        population_prop_obj = self.propDict["population"]
        clustered_by_property = PlotProperty.cluster_property(population_prop_obj)
        for property, properties in clustered_by_property.items():
            timeseries = [prop.get_timeseries_obj() for prop in properties]
            populations = [prop.population for prop in properties]
            for ts, population in zip(timeseries, populations):
                self.populationPlots.append(PopulationPlot(self, property, ts, population))
    
        dockName = f"population plots"
        newDock = Dock(dockName, closable = True)
        dockLayout = DockArea()
        newDock.addWidget(dockLayout)
        self.area.addDock(newDock, "bottom")

        for i, plot in enumerate(self.populationPlots):
            if i == 0:
                position = "bottom"
                relativeTo = None
            else:
                position = "above"
                plotRelativeTo = self.populationPlots[i-1]
                relativeTo = self.docks[plotRelativeTo]

            name = f"{plot.property}"
            plotDock = Dock(name, closable = True)
            plotDock.addWidget(plot.getPlotWidget())
            self.docks[plot] = (plotDock)

            dockLayout.addDock(plotDock, position, relativeTo)
    
    def addMplPlots():
        pass

    def updateSingleCellPlots(self):
        for plot in self.singleCellPlots:
            plot.update()
        

    def updatePopulationPlots(self):
        for plot in self.populationPlots:
            plot.update()
    
    def updateHeatMaps(self):
        for plot in self.heatMaps:
            plot.update()

    def setData(self):
        self.data = self.parent.getCellData()

    def getDataIdx(self, population):
        try:
            idx = self.parent.labelSelect.items().index(population)
            print(idx)
            return idx
        except ValueError:
            return None

    def updatePlots(self):
        self.updateSingleCellPlots()
        self.updatePopulationPlots()
        self.updateHeatMaps()

    def createTable(self):
        # create table with dummy data
        self.table = QtWidgets.QTableView()
        self.model = TableModel(self.parent.getCellDataAbbrev(), self.parent)
        self.table.setModel(self.model)

        self.table.selectionModel().currentChanged.connect(self.model.handleselect)

        # configure the table appearance
        self.table.verticalHeader().setVisible(False) # turn off vertical header
        self.table.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch) # stretch last section to take full width of allottedd area
        self.table.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.table.resizeColumnsToContents()
        self.hLayout.addWidget(self.table, 4)
        
    def resizeTableToContents(self):
        vh = self.table.verticalHeader()
        hh = self.table.horizontalHeader()
        size = QtCore.QSize(hh.length(), vh.length())  # Get the length of the headers along each axis.
        size += QtCore.QSize(vh.size().width(), hh.size().height())  # Add on the lengths from the *other* header
        size += QtCore.QSize(20, 20)  # Extend further so scrollbars aren't shown.
        self.resize(size)
    
    def closeEvent(self, event):
        self.parent.plotButton.setCheckState(False)


import torch
import torch.nn as nn
import torch.nn.functional as F
from ..rife_model.warplayer import warp
from ..rife_model.refine import *

def deconv(in_planes, out_planes, kernel_size=4, stride=2, padding=1):
    return nn.Sequential(
        torch.nn.ConvTranspose2d(in_channels=in_planes, out_channels=out_planes, kernel_size=4, stride=2, padding=1),
        nn.PReLU(out_planes)
    )

def conv(in_planes, out_planes, kernel_size=3, stride=1, padding=1, dilation=1):
    return nn.Sequential(
        nn.Conv2d(in_planes, out_planes, kernel_size=kernel_size, stride=stride,
                  padding=padding, dilation=dilation, bias=True),
        nn.PReLU(out_planes)
    )

class IFBlock(nn.Module):
    def __init__(self, in_planes, c=64):
        super(IFBlock, self).__init__()
        self.conv0 = nn.Sequential(
            conv(in_planes, c//2, 3, 2, 1),
            conv(c//2, c, 3, 2, 1),
            )
        self.convblock = nn.Sequential(
            conv(c, c),
            conv(c, c),
            conv(c, c),
            conv(c, c),
            conv(c, c),
            conv(c, c),
            conv(c, c),
            conv(c, c),
        )
        self.lastconv = nn.ConvTranspose2d(c, 5, 4, 2, 1)

    def forward(self, x, flow, scale):
        if scale != 1:
            x = F.interpolate(x, scale_factor = 1. / scale, mode="bilinear", align_corners=False)
        if flow != None:
            flow = F.interpolate(flow, scale_factor = 1. / scale, mode="bilinear", align_corners=False) * 1. / scale
            x = torch.cat((x, flow), 1)
        x = self.conv0(x)
        x = self.convblock(x) + x
        tmp = self.lastconv(x)
        tmp = F.interpolate(tmp, scale_factor = scale * 2, mode="bilinear", align_corners=False)
        flow = tmp[:, :4] * scale * 2
        mask = tmp[:, 4:5]
        return flow, mask
    
class IFNet(nn.Module):
    def __init__(self):
        super(IFNet, self).__init__()
        self.block0 = IFBlock(6, c=240)
        self.block1 = IFBlock(13+4, c=150)
        self.block2 = IFBlock(13+4, c=90)
        self.block_tea = IFBlock(16+4, c=90)
        self.contextnet = Contextnet()
        self.unet = Unet()

    def forward(self, x, scale=[4,2,1], timestep=0.5):
        img0 = x[:, :3]
        img1 = x[:, 3:6]
        gt = x[:, 6:] # In inference time, gt is None
        flow_list = []
        merged = []
        mask_list = []
        warped_img0 = img0
        warped_img1 = img1
        flow = None 
        loss_distill = 0
        stu = [self.block0, self.block1, self.block2]
        for i in range(3):
            if flow != None:
                flow_d, mask_d = stu[i](torch.cat((img0, img1, warped_img0, warped_img1, mask), 1), flow, scale=scale[i])
                flow = flow + flow_d
                mask = mask + mask_d
            else:
                flow, mask = stu[i](torch.cat((img0, img1), 1), None, scale=scale[i])
            mask_list.append(torch.sigmoid(mask))
            flow_list.append(flow)
            warped_img0 = warp(img0, flow[:, :2])
            warped_img1 = warp(img1, flow[:, 2:4])
            merged_student = (warped_img0, warped_img1)
            merged.append(merged_student)
        if gt.shape[1] == 3:
            flow_d, mask_d = self.block_tea(torch.cat((img0, img1, warped_img0, warped_img1, mask, gt), 1), flow, scale=1)
            flow_teacher = flow + flow_d
            warped_img0_teacher = warp(img0, flow_teacher[:, :2])
            warped_img1_teacher = warp(img1, flow_teacher[:, 2:4])
            mask_teacher = torch.sigmoid(mask + mask_d)
            merged_teacher = warped_img0_teacher * mask_teacher + warped_img1_teacher * (1 - mask_teacher)
        else:
            flow_teacher = None
            merged_teacher = None
        for i in range(3):
            merged[i] = merged[i][0] * mask_list[i] + merged[i][1] * (1 - mask_list[i])
            if gt.shape[1] == 3:
                loss_mask = ((merged[i] - gt).abs().mean(1, True) > (merged_teacher - gt).abs().mean(1, True) + 0.01).float().detach()
                loss_distill += (((flow_teacher.detach() - flow_list[i]) ** 2).mean(1, True) ** 0.5 * loss_mask).mean()
        c0 = self.contextnet(img0, flow[:, :2])
        c1 = self.contextnet(img1, flow[:, 2:4])
        tmp = self.unet(img0, img1, warped_img0, warped_img1, mask, flow, c0, c1)
        res = tmp[:, :3] * 2 - 1
        merged[2] = torch.clamp(merged[2] + res, 0, 1)
        return flow_list, mask_list[2], merged, flow_teacher, merged_teacher, loss_distill



import cv2
from cv2 import merge, resize
import numpy as np
from yeastvision.models.cp import CustomCPWrapper
from skimage.measure import label
from yeastvision.models.utils import addMasks
from yeastvision.models.artilife.budSeg.model import BudSeg
from skimage.measure import label
from skimage.morphology import remove_small_objects
from yeastvision.track.track import trackYeasts
from yeastvision.track.mat import get_mating_data
from yeastvision.track.cell import Cell, getBirthFrame, getCellData, getDeathFrame, getLifeData
from yeastvision.models.utils import normalizeIm, produce_weight_path
import torch



class Artilife(CustomCPWrapper):
    hyperparams  = { 
    "Mean Diameter":30, 
    "Flow Threshold":0.4, 
    "Cell Probability Threshold": 0,
    "Time Series": False}
    types = [None, None,None, bool]

    '''
    params also include:
        matStart - start index of mating cells
        matStop - stop index of mating cells (inclusive)
        matSeg - weights to use for matSeg 

        sporeStart - start index of sporulating cells
        sporeStop - stop index of sporulating cells (inclusive)
        tetradSeg - weights to use for tetradSeg
    these are hard coded via guiparts.ArtilifeParamDialog
    '''
    def __init__(self, params, weights):
        super().__init__(params, weights)

        if params["Time Series"]:
            self.budSeg = BudSeg()

    def findSmallBud(self, ims, masks):
        smallBuds = []
        for imOrig, mask in zip(ims,masks):
            im = resize((normalizeIm(imOrig)*255).astype(np.uint8), (160,160))
            im = merge((im,im,im))
            pred = self.budSeg.model.predict(np.expand_dims(im,0))
            pred = resize(pred[0,:,:,1],(50,50))
            pred = (pred==1).astype(np.uint8)
            r,c = pred.shape
            pred = np.pad(pred[10:r-10,10:c-10], 10)
            pred[mask>0] = 0
            smallBud = self.keepLargestObject(pred)
            smallBuds.append(smallBud)
        return np.array(smallBuds, dtype = np.uint8)
    
    def keepLargestObject(self, mask):
        labeled = label(mask.copy())
        labeled = label(remove_small_objects(labeled, 5))
        areas = []
        for num in np.unique(labeled):
            if num>0:
                areas.append((labeled==num).sum())
        if areas:
            cellToKeep = areas.index(max(areas))+1
            return (labeled==cellToKeep).astype(np.uint8)
        else:
            return np.zeros_like(mask)

    def addSmallCells(self, ims, cellData):      
        newMasks = self.masks.copy()
        imsPadded = np.array([np.pad(im, 25, mode = "symmetric") for im in ims], dtype = np.uint16)
        masksPadded = np.array([np.pad(mask, 25) for mask in newMasks], dtype = np.uint16)
        
        for val in range(len(cellData["birth"])):
            cellVal = int(val)+1
            if cellVal>0:
                birthFrame = cellData["birth"][cellVal-1]
                
                if birthFrame > 0:
                    cell = Cell(masksPadded[birthFrame,:,:], cellVal)
                    r,c = cell.centroid
                    r,c = round(r), round(c)
                    index = 2 if birthFrame>1 else 1
                    budSlice = slice(birthFrame-index,birthFrame), slice(r-25,r+25), slice(c-25,c+25)
                    smallBuds = self.findSmallBud(imsPadded[budSlice], masksPadded[budSlice])
                    if np.any(smallBuds>0):
                        birth = getBirthFrame(smallBuds, 1)
                        masksPadded[budSlice][smallBuds>0] = cellVal
        r,c = masksPadded[0].shape
        self.masks = masksPadded[:,25:r-25,25:c-25].astype(np.uint16)

    @classmethod
    def run(cls, ims, params, weights):
        params = params if params else cls.hyperparams
        model = cls(params, weights)
        ims3D = [cv2.merge((im,im,im)) for im in ims]
        assert len(ims3D[0].shape)==3

        if not params["Mean Diameter"]:
            evaluator = model.cpAlone
            model.masks, flows, _ = evaluator.eval(ims3D, 
                                                    diameter = model.params["Mean Diameter"], 
                                                    channels = [0, 0],
                                                    cellprob_threshold = model.params["Flow Threshold"], 
                                                    do_3D=False)
        else:
            evaluator = model.model
    
            model.masks, flows, _, model.diams = evaluator.eval(ims3D, 
                                                    diameter = model.params["Mean Diameter"], 
                                                    channels = [0, 0],
                                                    cellprob_threshold = model.params["Flow Threshold"], 
                                                    do_3D=False)
        print("process probability")
        model.cellprobs = [flow[2] for flow in flows]
        model.cellprobs = np.array((model.processProbability(model.cellprobs)), dtype = np.uint8)
        
        print("formatting return")
        model.masks = np.array(model.masks, dtype = np.uint16)

        if model.params["Time Series"]:
            model.masks = trackYeasts(model.masks)
            cellData = getLifeData(model.masks)
            model.addSmallCells(ims, cellData)
            model.masks = trackYeasts(model.masks)

        print("finished")
        mask, probs = model.masks, model.cellprobs

        del model
        
        return mask, probs



class ArtilifeFullLifeCycle(Artilife):
    hyperparams  = { 
    "Mean Diameter":30, 
    "Flow Threshold":0.4, 
    "Cell Probability Threshold": 0,
    "Mating Cells": False,
    "Sporulating Cells": False,
    "Time Series": False}
    types = [None, None,None, bool, bool, bool, bool]

    '''
    params also include:
        matStart - start index of mating cells
        matStop - stop index of mating cells (inclusive)
        matSeg - weights to use for matSeg 

        sporeStart - start index of sporulating cells
        sporeStop - stop index of sporulating cells (inclusive)
        tetradSeg - weights to use for tetradSeg
    these are hard coded via guiparts.ArtilifeParamDialog
    '''
    def __init__(self, params, weights):
        super().__init__(params, weights)

        self.matSeg, self.tetradSeg = None,None
        self.matMasks, self.matprobs = None, None
        self.tetraMasks, self.tetraprobs = None, None

        if params["Time Series"]:
            self.budSeg = BudSeg()
        if params["Mating Cells"]:
            from models.matSeg.model import MatSeg
            self.matSeg = MatSeg(self.params, produce_weight_path("matSeg", self.params["matSeg"]))

        if params["Sporulating Cells"]:
            from models.tetradSeg.model import TetradSeg
            self.tetradSeg = TetradSeg(self.params, produce_weight_path("tetradSeg", self.params["tetradSeg"]))

    def addTetrads(self, ims):
        tetraSlice = slice(int(self.params["sporeStart"]), int(self.params["sporeStop"])+1)
        self.tetraMasks, tetraFlows, _, _ = self.tetradSeg.model.eval(ims[tetraSlice], 
                                                                diameter = self.params["Mean Diameter"], 
                                                                channels = [0,0],
                                                                cellprob_threshold = self.params["Flow Threshold"],
                                                                do_3D = False)
        self.tetraprobs = [flow[2] for flow in tetraFlows]
        self.tetraprobs = np.array(self.processProbability(self.tetraprobs), dtype = np.uint8)

        if self.params["Time Series"]:
            newTetraMasks, newMasks = track_obj([im[:,:,0] for im in ims[tetraSlice]], self.masks[tetraSlice], self.tetraMasks[tetraSlice], False)
            tetraSlice = slice(int(self.params["sporeStart"]), int(self.params["sporeStop"])-1)
        else:
            for tetraMask, cellMask in zip(self.tetraMasks, self.masks):
                if np.any(tetraMask>=1):
                    newMask = addMasks(tetraMask, cellMask)
                    newMasks.append(newMask)
                else:
                    newMasks.append(cellMask)
            newTetraMasks = self.tetraMasks

        self.masks[tetraSlice] = newMasks
        tetraMasks = np.zeros_like(self.masks)
        tetraMasks[tetraSlice] = np.array(newTetraMasks, dtype = np.uint16)
        self.tetraMasks = tetraMasks


    def addMatingCells(self, ims):
        matSlice = slice(int(self.params["matStart"]),int(self.params["matStop"]) + 1)
        self.matMasks, matFlows, _, _ = self.matSeg.model.eval(ims[matSlice], 
                                                                diameter = self.params["Mean Diameter"], 
                                                                channels = [0,0],
                                                                cellprob_threshold = self.params["Flow Threshold"],
                                                                do_3D = False)
        print("got mating cells")
        self.matprobs = [flow[2] for flow in matFlows]
        self.matprobs = np.array(self.processProbability(self.matprobs), dtype = np.uint8)
        
        if self.params["Time Series"]:
            matSlice = slice(int(self.params["matStart"]), int(self.params["matStop"])-1)
            newMatMasks, newMasks = get_mating_data(self.matMasks[matSlice], self.masks[matSlice])
        else:
            newMasks = []
            for matMask, cellMask in zip(self.matMasks, self.masks):
                if np.any(matMask>=1):
                    newMask = addMasks(matMask, cellMask)
                    newMasks.append(newMask)
                else:
                    newMasks.append(cellMask)
            newMatMasks = self.matMasks

        self.masks[matSlice] = newMasks
        matMasks = np.zeros_like(self.masks)
        matMasks[matSlice] = np.array(newMatMasks, dtype = np.uint16)
        self.matMasks = matMasks

    def findSmallBud(self, ims, masks):
        smallBuds = []
        for imOrig, mask in zip(ims,masks):
            im = resize((normalizeIm(imOrig)*255).astype(np.uint8), (160,160))
            im = merge((im,im,im))
            pred = self.budSeg.model.predict(np.expand_dims(im,0))
            pred = resize(pred[0,:,:,1],(50,50))
            pred = (pred==1).astype(np.uint8)
            r,c = pred.shape
            pred = np.pad(pred[10:r-10,10:c-10], 10)
            pred[mask>0] = 0
            smallBud = self.keepLargestObject(pred)
            smallBuds.append(smallBud)
        return np.array(smallBuds, dtype = np.uint8)
    
    def keepLargestObject(self, mask):
        labeled = label(mask.copy())
        labeled = label(remove_small_objects(labeled, 5))
        areas = []
        for num in np.unique(labeled):
            if num>0:
                areas.append((labeled==num).sum())
        if areas:
            cellToKeep = areas.index(max(areas))+1
            return (labeled==cellToKeep).astype(np.uint8)
        else:
            return np.zeros_like(mask)

    def addSmallCells(self, ims, cellData):      
        newMasks = self.masks.copy()
        imsPadded = np.array([np.pad(im, 25, mode = "symmetric") for im in ims], dtype = np.uint16)
        masksPadded = np.array([np.pad(mask, 25) for mask in newMasks], dtype = np.uint16)
        
        for val in range(len(cellData["birth"])):
            cellVal = int(val)+1
            if cellVal>0:
                birthFrame = cellData["birth"][cellVal-1]
                
                if birthFrame > 0:
                    cell = Cell(masksPadded[birthFrame,:,:], cellVal)
                    r,c = cell.centroid
                    r,c = round(r), round(c)
                    index = 2 if birthFrame>1 else 1
                    budSlice = slice(birthFrame-index,birthFrame), slice(r-25,r+25), slice(c-25,c+25)
                    smallBuds = self.findSmallBud(imsPadded[budSlice], masksPadded[budSlice])
                    if np.any(smallBuds>0):
                        birth = getBirthFrame(smallBuds, 1)
                        masksPadded[budSlice][smallBuds>0] = cellVal
        r,c = masksPadded[0].shape
        self.masks = masksPadded[:,25:r-25,25:c-25].astype(np.uint16)

    @classmethod
    @torch.no_grad()
    def run(cls, ims, params, weights):
        params = params if params else cls.hyperparams
        model = cls(params, weights)
        ims3D = [cv2.merge((im,im,im)) for im in ims]
        assert len(ims3D[0].shape)==3

        if not params["Mean Diameter"]:
            evaluator = model.cpAlone
            model.masks, flows, _ = evaluator.eval(ims3D, 
                                                    diameter = model.params["Mean Diameter"], 
                                                    channels = [0, 0],
                                                    cellprob_threshold = model.params["Flow Threshold"], 
                                                    do_3D=False)
        else:
            evaluator = model.model
            model.masks, flows, _, model.diams = evaluator.eval(ims3D, 
                                                    diameter = model.params["Mean Diameter"], 
                                                    channels = [0, 0],
                                                    cellprob_threshold = model.params["Flow Threshold"], 
                                                    do_3D=False)
        print("process probability")
        model.cellprobs = [flow[2] for flow in flows]
        model.cellprobs = np.array((model.processProbability(model.cellprobs)), dtype = np.uint8)
        
        print("formatting return")
        model.masks = np.array(model.masks, dtype = np.uint16)

        if model.params["Time Series"]:
            model.masks = trackYeasts(model.masks)
            cellData = getLifeData(model.masks)
            model.addSmallCells(ims, cellData)
            model.masks = trackYeasts(model.masks)
            del model.budSeg

        if model.matSeg:
            print('add mating cell')
            model.addMatingCells(ims3D)
        if model.tetradSeg:
            model.addTetrads(ims3D)
        
        print("finished")
        arti = (model.masks, model.cellprobs)
        mating = (model.matMasks, model.matprobs)
        tetra = (model.tetraMasks, model.tetraprobs)

        del modelf

        return {"artilife": arti,
                "mating": mating,
                "tetrads": tetra
                }


    



from PyQt5.QtWidgets import QAction
from yeastvision.disk import io

def menubar(parent):
    mainMenu = parent.menuBar()
    fileMenu = mainMenu.addMenu("File")

    loadIm = QAction("Load Experiment", parent)
    loadIm.triggered.connect(parent.userSelectExperiment)
    fileMenu.addAction(loadIm)

    editMenu = mainMenu.addMenu("Edit")
    addBlankMask = QAction("Add Blank Label", parent)
    addBlankMask.triggered.connect(parent.addBlankMasks)
    editMenu.addAction(addBlankMask)
    removeAll = QAction("Clear All", parent)
    removeAll.triggered.connect(parent.setEmptyDisplay)
    clearCurrent = QAction("Clear current mask", parent)
    clearCurrent.triggered.connect(parent.clearCurrMask)
    editMenu.addAction(clearCurrent)
    editMenu.addAction(removeAll)
    labelObjects = QAction("Label Unique Regions", parent)
    labelObjects.triggered.connect(parent.labelCurrMask)
    editMenu.addAction(labelObjects)

    preMenu = mainMenu.addMenu("Preprocess")
    medianFilter = QAction("Median Filter", parent)
    medianFilter.triggered.connect(parent.doMedian)
    preMenu.addAction(medianFilter)
    gaussFilter = QAction("Gaussian Filter", parent)
    gaussFilter.triggered.connect(parent.doGaussian)
    preMenu.addAction(gaussFilter)
    adaptHist = QAction("Adaptive Histogram Equalization", parent)
    adaptHist.triggered.connect(parent.doAdaptHist)
    preMenu.addAction(adaptHist)
    rescale = QAction("Rescale", parent)
    # cannot add a threadd for rescale because of error with parent from other thread
    rescale.triggered.connect(lambda: parent.doRescale())
    preMenu.addAction(rescale)

    exportMenu = mainMenu.addMenu("Export")
    imSave = QAction("Save Images", parent)
    imSave.triggered.connect(parent.saveIms)
    exportMenu.addAction(imSave)
    maskSave = QAction("Save Masks", parent)
    maskSave.triggered.connect(parent.saveMasks)
    exportMenu.addAction(maskSave)
    overlaySave = QAction("Save Figure", parent)
    overlaySave.triggered.connect(parent.saveFigure)
    exportMenu.addAction(overlaySave)
    dataSave = QAction("Export Cell Data", parent)
    dataSave.triggered.connect(parent.saveCellData)
    exportMenu.addAction(dataSave)
    hmSave = QAction("Export Heatmaps", parent)
    hmSave.triggered.connect(parent.saveHeatMaps)
    exportMenu.addAction(hmSave)
    lineageSave = QAction("Export Lineage Data", parent)
    lineageSave.triggered.connect(parent.saveLineageData)
    exportMenu.addAction(lineageSave)
    figureCreate = QAction("Create Label Overlay", parent)
    figureCreate.triggered.connect(parent.createMaskOverlay)
    exportMenu.addAction(figureCreate)

    modelMenu = mainMenu.addMenu("Models")
    trainModel = QAction("Train Model", parent)
    trainModel.triggered.connect(parent.showTW)
    modelMenu.addAction(trainModel)

    loadCustom = QAction("Load Custom Model", parent)
    loadCustom.triggered.connect(lambda: io.loadCustomModel(parent))
    modelMenu.addAction(loadCustom)

    evaluate = QAction("Evaluate Predictions", parent)
    evaluate.triggered.connect(parent.evaluate)
    modelMenu.addAction(evaluate)

    plotMenu = mainMenu.addMenu("Plot")

    getCellData = QAction("Update Cell Data", parent)
    getCellData.triggered.connect(parent.updateCellData)
    plotMenu.addAction(getCellData)


    


import tensorflow.keras.backend as K
import tensorflow as tf 


def jaccard_coef(y_true, y_pred):
    y_true_f = K.flatten(y_true)
    y_pred_f = K.flatten(y_pred)
    intersection = K.sum(y_true_f * y_pred_f)
    return (intersection + 1.0) / (K.sum(y_true_f) + K.sum(y_pred_f) - intersection + 1.0)

def jaccard_coef_loss(y_true, y_pred):
    return -jaccard_coef(y_true, y_pred)


import torch
import torch.nn as nn
import numpy as np
from torch.optim import AdamW
import torch.optim as optim
import itertools
from ..rife_model.warplayer import warp
from torch.nn.parallel import DistributedDataParallel as DDP
from ..rife_model.IFNet import *
from ..rife_model.IFNet_m import *
import torch.nn.functional as F
from ..rife_model.loss import *
from ..rife_model.laplacian import *
from ..rife_model.refine import *

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    
class Model:
    def __init__(self, local_rank=-1, arbitrary=False):
        if arbitrary == True:
            self.flownet = IFNet_m()
        else:
            self.flownet = IFNet()
        self.device()
        self.optimG = AdamW(self.flownet.parameters(), lr=1e-6, weight_decay=1e-3) # use large weight decay may avoid NaN loss
        self.epe = EPE()
        self.lap = LapLoss()
        self.sobel = SOBEL()
        if local_rank != -1:
            self.flownet = DDP(self.flownet, device_ids=[local_rank], output_device=local_rank)

    def train(self):
        self.flownet.train()

    def eval(self):
        self.flownet.eval()

    def device(self):
        self.flownet.to(device)

    def load_model(self, path, rank=0):
        def convert(param):
            return {
            k.replace("module.", ""): v
                for k, v in param.items()
                if "module." in k
            }
            
        if rank <= 0:
            self.flownet.load_state_dict(convert(torch.load('{}/flownet.pkl'.format(path), map_location=device)))
        
    def save_model(self, path, rank=0):
        if rank == 0:
            torch.save(self.flownet.state_dict(),'{}/flownet.pkl'.format(path))

    def inference(self, img0, img1, scale=1, scale_list=[4, 2, 1], TTA=False, timestep=0.5):
        for i in range(3):
            scale_list[i] = scale_list[i] * 1.0 / scale
        imgs = torch.cat((img0, img1), 1)
        flow, mask, merged, flow_teacher, merged_teacher, loss_distill = self.flownet(imgs, scale_list, timestep=timestep)
        if TTA == False:
            return merged[2]
        else:
            flow2, mask2, merged2, flow_teacher2, merged_teacher2, loss_distill2 = self.flownet(imgs.flip(2).flip(3), scale_list, timestep=timestep)
            return (merged[2] + merged2[2].flip(2).flip(3)) / 2
    
    def update(self, imgs, gt, learning_rate=0, mul=1, training=True, flow_gt=None):
        for param_group in self.optimG.param_groups:
            param_group['lr'] = learning_rate
        img0 = imgs[:, :3]
        img1 = imgs[:, 3:]
        if training:
            self.train()
        else:
            self.eval()
        flow, mask, merged, flow_teacher, merged_teacher, loss_distill = self.flownet(torch.cat((imgs, gt), 1), scale=[4, 2, 1])
        loss_l1 = (self.lap(merged[2], gt)).mean()
        loss_tea = (self.lap(merged_teacher, gt)).mean()
        if training:
            self.optimG.zero_grad()
            loss_G = loss_l1 + loss_tea + loss_distill * 0.01 # when training RIFEm, the weight of loss_distill should be 0.005 or 0.002
            loss_G.backward()
            self.optimG.step()
        else:
            flow_teacher = flow[2]
        return merged[2], {
            'merged_tea': merged_teacher,
            'mask': mask,
            'mask_tea': mask,
            'flow': flow[2][:, :2],
            'flow_tea': flow_teacher,
            'loss_l1': loss_l1,
            'loss_tea': loss_tea,
            'loss_distill': loss_distill,
            }
